#include "emu.h"
#include "global.h"
#include <ddraw.h>
#include "input.h"
#include "gui.h"
#include "direct3d.h"
#include "Memory.h"

#include "snes9x/port.h"
#include "snes9x/memmap.h"
#include "snes9x/display.h"
#include "snes9x/apu.h"
#include "snes9x/cpuexec.h"
#include "snes9x/soundux.h"
#include "snes9x/gfx.h"
#include "snes9x/debug.h"

bool enabled = false;

static LPDIRECTDRAWSURFACE4 screen = NULL;
static LPDIRECTDRAWSURFACE4 subscreen = NULL;

static unsigned short *GFXscreen = NULL;
static unsigned short *GFXsubscreen = NULL;
static unsigned char *GFXzbuffer = NULL;
static unsigned char *GFXsubzbuffer = NULL;

static FILE* td = NULL;

extern void S9xSyncSpeed() {
	// Frame skip algorithm - very similar to a line-drawing algorithm
	// Based on the idea of a line connecting the origin (0, 0) and (skip, outof)
	// A frame is skipped once for every x, at the closest y location (frame)
	// For example: skip = 3, outof = 8
	// 3 frames out of 8 are skipped.
	// Line equation: y = mx + b, where m = 8/3 and b = 0
	// At x = 1, y = (8/3)1 + 1 = 8/3 = 2 2/3, or approximately 3 (rounded)
	// The third frame will be skipped.
	// 
	// Evaluation of this algorithm is as follows:
	// y = mx + b, b = 0, m = ^y/^x    (^x = delta x, or skip; ^y = delta y, or outof)
	// y = (^y/^x)x
	// x = (^x/^y)y
	// For every y, if x is greater than or equal to the last skipped frame + 1/2,
	//  then the current y location (this frame) is skipped.
	// ^x and ^y do not change through all iterations of y (except in special cases)
	// Thus, for every y, x changes only by y multiplied by 1/m, 
	// This means, beginning at 0, 1/m may be added to x for each iteration of y, solving x
	// Since 1/m = dx/dy, which is usually a decimal, it is more convient to use integer math
	// Since dy x = dx y, and dy never changes, it is easier to solve for ^y x rather than x
	// Thus, to solve for dy x, we simple add dx for each iteration of y.
	// If x is greater than the last skipped frame + 1/2, then this frame should be skipped
	//  (this is stated above)
	// Thus, if x > 1/2, we skip
	// Multiplying both sides by dy, we see:
	// if dy x > dy / 2, skip
	// That is to say, if dy x (the solution to the equation) is greater than dy/2, we should skip
	// Again, multiplying both sides by 2, we see:
	// if 2(dy x) > dy, skip
	// This eliminates fractions.
	// Moving backwards, 2(dy x) = 2(dx y)
	// Thus, accumulating 2(dx) solves for 2(dy x)
	// To prevent 2(dy x) from accumulating ad infinitum, 2(dy) should be subtracted from 2(dy x)
	//  every time a frame is skipped.
	//
	// In the algorithm below:
	// d = 2(dy x), the accumulation of 2(dx), a decision variable
	// incrE = increment per y
	// decrNE = decrement if frame is skipped
	//
	// if(x2 >= outof) { skip; x2 -= outof2; }

	static int d = 0;
	static int dx, dy;
	
	// If no frames are to be skipped, render this frame and move on
	if(gint_skipframes == 0) {
		IPPU.SkippedFrames = 0;
		IPPU.RenderThisFrame = TRUE;
		return;
	}

	// If for some reason, the frameskip has changed, reset relevant variables
	if(dx != gint_skipframes) {
		d = 0;
	}		

	// Setup dx and dy
	dx = gint_skipframes;
	dy = gint_skipoutof;
	int incrE = 2 * dx;
	int decrNE = 2 * dy;
	
	// Accumulate
	d += incrE;

	// Check for bounds
	if(d >= dy) {
		IPPU.SkippedFrames++;
		IPPU.RenderThisFrame = FALSE;
		d -= decrNE;
	} else {
		IPPU.SkippedFrames = 0;
		IPPU.RenderThisFrame = TRUE;
	} 

}

//extern TCHAR* superstr;

void DoSnes9xInit(void) {
	ZeroMemory(&Settings, sizeof(Settings));

	Settings.SoundPlaybackRate = 4;
	Settings.Stereo = TRUE;
	Settings.SoundBufferSize = 0; 
	Settings.DisableSoundEcho = FALSE; 
	Settings.DisableSampleCaching = FALSE;
	Settings.DisableMasterVolume = FALSE;
	Settings.ThreadSound = FALSE;

	Settings.SPCTo65c816Ratio = 2;
	Settings.APUEnabled = TRUE;
	Settings.H_Max = (100 * SNES_CYCLES_PER_SCANLINE) / 100;
	Settings.SkipFrames = AUTO_FRAMERATE;
	Settings.Shutdown = TRUE;
	Settings.FrameTimePAL = 20000;
	Settings.FrameTimeNTSC = 16667;
	Settings.FrameTime = Settings.FrameTimeNTSC;
	Settings.Mouse = FALSE;
	Settings.SuperScope = FALSE; //TRUE;
	Settings.ControllerOption = SNES_MULTIPLAYER5;
	Settings.Transparency = TRUE;
	Settings.SixteenBit = TRUE;
	Settings.SupportHiRes = FALSE; //TRUE;
	Settings.NetPlay = FALSE;
	Settings.ServerName[0] = 0;
	Settings.AutoSaveDelay = 30;
//	Settings.TraceUnknownRegisters = TRUE;
//	Settings.DisplayFrameRate = TRUE;

	if(Settings.Transparency) Settings.SixteenBit = TRUE;

	Settings.HBlankStart = (256 * Settings.H_Max) / SNES_HCOUNTER_MAX;

	DoMemoryManager(_T("Before Snes9x Init"));

	if(!Memory.Init()){
		S9xMessage(0, 0, "Out of Memory, Memory.Init()");
	}

	DoMemoryManager(_T("After Memory.Init()"));

	if(!S9xInitAPU()) {
		S9xMessage(0, 0, "Out of Memory, S9xInitAPU()");
	}

	DoMemoryManager(_T("After S9xInitAPU()"));


/*#if defined UNDER_CE
	const char file[] = "\\CD-ROM\\CHRONO.SMC";
#else
	const char file[] = "CHRONO.SMC";
#endif

	if(!Memory.LoadROM(file)) {
#if defined USE_DIRECT_X
		n = wsprintf(str, TEXT("Failed to load ROM"));
		ExtTextOut(dc, 0, 32, NULL, NULL, str, n, NULL);
#endif
	} */

#if defined USE_DIRECT_X
	DDSURFACEDESC2			ddsd;
	HRESULT					hr;

	memset(&ddsd, 0, sizeof(ddsd));
	gdds_screen->GetSurfaceDesc(&ddsd);  

	ddsd.dwSize         = sizeof(DDSURFACEDESC2);
	ddsd.dwFlags        = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
	ddsd.ddsCaps.dwCaps = DDSCAPS_VIDEOMEMORY;

	ddsd.dwWidth        = 256;
	ddsd.dwHeight       = 239;
	hr = gpdd_directdraw->CreateSurface(&ddsd, &screen, NULL);
	if(hr != DD_OK) {
		S9xMessage(0, 0, "Failed to create DirectDraw screen surface");
	}

	DoMemoryManager(_T("After CreateSurface()"));

#if defined RENDER_DIRECT_TO_VIDEO_MEMORY
	hr = gpdd_directdraw->CreateSurface(&ddsd, &subscreen, NULL);

	if(hr != DD_OK) {
		S9xMessage(0, 0, "Failed to create DirectDraw subscreen surface");
	}
#endif
#endif

	GFXzbuffer = new unsigned char[256*239];
	GFXsubzbuffer = new unsigned char[256*239];
	GFX.Pitch = 256*2;
	GFX.ZBuffer = GFXzbuffer;
	GFX.SubZBuffer = GFXsubzbuffer;

	DoMemoryManager(_T("After new zbuffer"));

#if !defined RENDER_DIRECT_TO_VIDEO_MEMORY
	GFXscreen = new unsigned short[256*239];
	GFXsubscreen = new unsigned short[256*239];

	DoMemoryManager(_T("After new screen"));

	GFX.Screen = (unsigned char *)GFXscreen;
	GFX.SubScreen = (unsigned char *)GFXsubscreen;
	S9xGraphicsInit();

	DoMemoryManager(_T("After S9xGraphicsInit()"));

	if(!GFXzbuffer || !GFXscreen || !GFXsubscreen || !GFXsubzbuffer) {
		S9xMessage(0, 0, "Out of Memory. Allocationg GFX buffers.");
	}
#endif

	S9xSetSoundMute(TRUE);

//	CPU.Flags |= DEBUG_MODE_FLAG;
}

void DoSnes9xEmulation(void) {
	MSG msg;

	while(1) {
		if(PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE)) {
			if(!GetMessage(&msg, NULL, 0, 0))
				return;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		CheckForDevices();
		S9xMainLoop();

//		for(int i = 0; i < 4; i++) {
//			int j = GetInputDeviceData(i);
//			DoDisplayInputData(i, j);
//		}
			
	}
}

bool8 S9xInitUpdate () {
	if(!IPPU.RenderThisFrame) return false;

#if defined RENDER_DIRECT_TO_VIDEO_MEMORY
	DDSURFACEDESC2	ddsd;

	ddsd.dwSize = sizeof(ddsd);
	screen->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);
	GFX.Screen = (unsigned char *)ddsd.lpSurface;

	subscreen->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);
	GFX.SubScreen = (unsigned char *)ddsd.lpSurface;

	S9xGraphicsInit();
#endif

	return true; 
}

bool8 S9xDeinitUpdate (int width, int height, bool8 sixteen_bit) {
	static int frame = 0;

	if(!IPPU.RenderThisFrame) return false;

#if defined RENDER_DIRECT_TO_VIDEO_MEMORY
	screen->Unlock(NULL);
	subscreen->Unlock(NULL);
	S9xGraphicsDeinit();
#else
#ifdef USE_DIRECT_X
	DDSURFACEDESC2	ddsd;

	ddsd.dwSize = sizeof(ddsd);
	screen->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);

	unsigned int *sp = (unsigned int *)ddsd.lpSurface;
	unsigned int *gp = (unsigned int *)GFXscreen;

	for(int y = 0; y < height; y++) {
		for(int x = 0; x < width/2; x++) {
//			if(x == 50) *(sp+x) = 0xFFFF;
//			else if(x == 100) *(sp+x) = frame % 63;
			/*else*/ *(sp+x) = *(gp+x);
		}
		sp += 128;
		gp += width/2;
	}

	screen->Unlock(NULL);
//	subscreen->Unlock(NULL);
#endif
#endif

	RECT rt = { 0, 0, width, height };
	RECT tvs = { 16, 12, 640-16, 480-12 };
	RECT tvl = { 32, 24, 640-32, 480-24 };
	RECT fs = { 0, 0, 640, 480 };

/*	TCHAR str[256];
	HDC dc;
	int n;
	screen->GetDC(&dc);
	SetTextColor(dc, RGB(255, 255, 255));
	SetBkMode(dc, TRANSPARENT);

	n = wsprintf(str, TEXT("%i x %i"), width, height);
	ExtTextOut(dc, 0, 0, NULL, NULL, str, n, NULL);

	screen->ReleaseDC(dc); */

#ifdef USE_DIRECT_X
	gdds_back->Blt(gint_videomode == 0 ? &tvs : 
		(gint_videomode == 1) ? &tvl : &fs, screen, &rt, DDBLT_WAIT, NULL);
	FlipScreen();
#endif

	frame++;
	return true; 
}

extern void S9xMessage(int type, int number, const char *message) {

#if !defined UNDER_CE
	static FILE *f = NULL;
	if(!f) {
		f = fopen("s9xmessage.trace.txt", "wt");
	}
	fprintf(f, "%s\n", message);
	fflush(f);
#endif

#if defined SHOW_S9XMESSAGE
#if defined USE_DIRECT_X
	static HFONT font = NULL;
	LOGFONT lf;
	if(!font) {
		memset(&lf, 0, sizeof(lf));
		lf.lfHeight = 32;
		font = CreateFontIndirect(&lf);
	}
	
	//	if(!enabled) return;

	DDBLTFX     ddbltfx;

    // Clear the back buffer
    memset(&ddbltfx, 0, sizeof(DDBLTFX));
    ddbltfx.dwSize = sizeof(DDBLTFX);
    ddbltfx.dwFillColor = 0;
    gdds_back->Blt(NULL, NULL, NULL, DDBLT_COLORFILL | DDBLT_WAIT, &ddbltfx);

	// Alright this code is here as overhead for showing text.
	// It's all over the place in my code so get over it
	// This code should not stay
	// NEW COMMENT: Most code like this has been removed and ported to a call to S9xMessage
	// This code may stay
	HDC				dc;
	TCHAR			str[1024];
	int n;

	gdds_back->GetDC(&dc);

	SelectObject(dc, font);

	SetTextColor(dc, RGB(255, 255, 255));
	SetBkMode(dc, TRANSPARENT);


#if defined UNDER_CE
	mbstowcs(str, message, strlen(message));
#else
	strcpy(str, message);
#endif

	n = strlen(message);
	ExtTextOut(dc, 50, 50, NULL, NULL, str, n, NULL);

	gdds_back->ReleaseDC(dc);
	gdds_screen->Flip(NULL, DDFLIP_WAIT);

#ifdef S9XMESSAGE_WAIT_FOR_INPUT
	int a = 0;
	while(!a) {
		for(int i = 0; i < 4; i++) {
			a |= GetInputDeviceData(i);
		}

		MSG msg;
		if(PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE)) {
			if(!GetMessage(&msg, NULL, 0, 0))
				return;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	// Wait until the button is depressed...
	while(a) {
		a = 0;
		for(int i = 0; i < 4; i++) {
			a |= GetInputDeviceData(i);
		}
	}
#else
	MSG msg;

	DWORD tick = GetTickCount();
	while(GetTickCount() < tick + 1000) {
		if(PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE)) {
			if(!GetMessage(&msg, NULL, 0, 0))
				return;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
#endif

#endif
#endif
}

extern uint32 S9xReadJoypad (int which1_0_to_4) { 
	int n = which1_0_to_4;
	if(n == 4) return 0;
	int ret = 0;

#if !defined UNDER_CE
	n = 2;
#endif

	int data = GetInputDeviceData(n);

#if !defined UNDER_CE
	if((data | BUTTON[0]) == data) {
		ret |= SNES_A_MASK;
	}
	if((data | BUTTON[1]) == data) {
		ret |= SNES_B_MASK;
	}
	if((data | BUTTON[2]) == data) {
		ret |= SNES_X_MASK;
	}
	if((data | BUTTON[3]) == data) {
		ret |= SNES_Y_MASK;
	}
	if((data | BUTTON[4]) == data) {
		ret |= SNES_START_MASK;
	}
	if((data | BUTTON[5]) == data) {
		ret |= SNES_SELECT_MASK;
	}
	if((data | UP) == data) {
		ret |= SNES_UP_MASK;
	}
	if((data | DOWN) == data) {
		ret |= SNES_DOWN_MASK;
	}
	if((data | RIGHT) == data) {
		ret |= SNES_RIGHT_MASK;
	}
	if((data | LEFT) == data) {
		ret |= SNES_LEFT_MASK;
	}
	if((data | BUTTON[6]) == data) {
		ret |= SNES_TL_MASK;
	}
	if((data | BUTTON[7]) == data) {
		ret |= SNES_TR_MASK;
	}

	if((data | BUTTON[8]) == data) {
			DoInGameMenu();
	}

#else
	if((data | BUTTON[2]) == data) {
		if((data | BUTTON[1]) == data) {
			ret |= SNES_START_MASK;
		}
		if((data | BUTTON[8]) == data) {
			ret |= SNES_SELECT_MASK;
		}
		if(((data | BUTTON[9]) == data) && ((data | BUTTON[10]) == data)) {
	//		S9xMessage(0, 0, "Hotkey received"); 
			DoInGameMenu();
		}
	}

	if((data | BUTTON[0]) == data) {
		ret |= SNES_A_MASK;
	}
	if((data | BUTTON[1]) == data) {
		if((data | BUTTON[2]) != data) {
			ret |= SNES_B_MASK;
		}
	}
	if((data | BUTTON[7]) == data) {
		ret |= SNES_X_MASK;
	}
	if((data | BUTTON[8]) == data) {
		if((data | BUTTON[2]) != data) {
			ret |= SNES_Y_MASK;
		}
	}
	if((data | UP) == data) {
		ret |= SNES_UP_MASK;
	}
	if((data | DOWN) == data) {
		ret |= SNES_DOWN_MASK;
	}
	if((data | RIGHT) == data) {
		ret |= SNES_RIGHT_MASK;
	}
	if((data | LEFT) == data) {
		ret |= SNES_LEFT_MASK;
	}
	if((data | BUTTON[3]) == data) {
		ret |= SNES_UP_MASK;
	}
	if((data | BUTTON[4]) == data) {
		ret |= SNES_DOWN_MASK;
	}
	if((data | BUTTON[6]) == data) {
		ret |= SNES_RIGHT_MASK;
	}
	if((data | BUTTON[5]) == data) {
		ret |= SNES_LEFT_MASK;
	}
	if((data | BUTTON[10]) == data) {
		ret |= SNES_TL_MASK;
	}
	if((data | BUTTON[9]) == data) {
		ret |= SNES_TR_MASK;
	}
#endif

	ret |= 0x80000000;

	return ret;

}


